#include "common.h"
#include "icmptunnel.h"



struct socket_struct	sockets[MAX_SOCKETS] ; 
struct sockaddr		peer ;


static int print_usage(const char *me) ;
static int print_version(void) ;
static int print_example(const char *me) ; 


struct {
   int listen_port ;
   int forward_port ;
   struct sockaddr forward_host ;
   char forward_host_ip[(3*4) + 3 + 1] ;	/* xxx * 4 + ... + \0 */
   char peer_ip[(3*4) + 3 + 1] ; 
   int recv_type ;
   int send_type ;
   int16_t tunnel_id ;
   int16_t tunnel_rid ;
   int opmode ; /* mode of operation .. LISTEN or FORWARD */
   int poll_mode ;
   int poll_interval ; 
   
} conf ;



int main(int argc, char *argv[])
{
   
   int s=0 ;
   int is=0 ;  /* icmp socket */
   int n=0 ;
   struct sockaddr_in *to ;

   
#ifdef GETOPTLONG
   int c = 0 ;
   int tempi = 0 ;
   char *szhostname ;
   char *szport ;
   struct hostent *hp;
#endif
   
   conf.listen_port = DEFAULT_LISTEN_PORT ;
   conf.forward_port = DEFAULT_FORWARD_PORT ;
   conf.recv_type = DEFAULT_RECV_TYPE ;
   conf.send_type = DEFAULT_SEND_TYPE ;
   conf.tunnel_id = DEFAULT_TUNNEL_ID ;
   conf.tunnel_rid = 0xFFFF ; 
   conf.opmode = DEFAULT_OPMODE ;
   conf.poll_mode = POLLMODE_OFF ;
   conf.poll_interval = DEFAULT_POLL_INTERVAL ; 
   
   
   
   
#ifdef GETOPTLONG
   
   while (1) {
      int option_index = 0 ; 
      static struct option long_options[] = 
	{
	     {"forward", required_argument, 0, 'F'},
	     {"listen", required_argument, 0, 'L'},
	     {"send_type", required_argument, 0, 'S'},
	     {"recv_type", required_argument, 0, 'R'},
	     {"id", required_argument, 0, 'I'},
	     {"rid", required_argument, 0, 'i'},
	     {"poll", required_argument, 0, 'P'},
	     {"list", no_argument, 0, 'l'}, 
	     {"version", no_argument, 0, 'V'},
	     {"help", no_argument, 0, 'h'},
	     {"example", no_argument, 0, 'e'}, 
	     {0, 0, 0, 0}
	} ;
      
      c = getopt_long(argc, argv, "F:L:S:R:I:i:P:Vhle", long_options, &option_index) ;
      if (c == EOF)
	break ; 
      
      switch(c) {
       case 0:
	 fprintf(stderr, "main:\t\t\toption %s", long_options[option_index].name) ;
	 
	 if (optarg) 
	    fprintf(stderr, " with arg %s", optarg) ;
	 fprintf(stderr, "\n") ;
	 
	 break ;
	 
       case 'e': /* --example | -e */
	 print_example(argv[0]) ;
	 exit(0) ; 
	 break ;
	 
       case 'F': /* --forward | -F hostname:port */
	 szport = (char *) malloc(strlen(optarg)) ;
	 sprintf(szport, "%s", optarg) ;
	 
	 szhostname = strsep(&szport, (char *) ":") ;
	 if (!szport)
	   conf.forward_port = DEFAULT_FORWARD_PORT ;
	 else
	   conf.forward_port = atoi(szport) ;
	 
	 to = (struct sockaddr_in *) &conf.forward_host ;
	    
	 if (!(inet_aton(szhostname, &to->sin_addr))) {
	    fprintf(stderr, "main:\t\t\tszhostname='%s'\n", szhostname) ; 
	    hp = gethostbyname(szhostname) ;
	    if (!hp) {
	       fprintf(stderr, "main:\t\t\tUnknown host: %s\n", szhostname) ;
	       exit(1) ;
	    }
	     
	    
	    to->sin_family = hp->h_addrtype ; 
	                    
	    if (hp->h_length > (int)sizeof(to->sin_addr)) {
	       hp->h_length = sizeof(to->sin_addr);
	    }
	    memcpy(&to->sin_addr, hp->h_addr, hp->h_length);
	 }
	 
	 sprintf(conf.forward_host_ip, "%s", inet_ntoa(to->sin_addr)) ;
	 conf.opmode = OPMODE_FORWARD ; 
	 
	 break ;

       case 'I': /* --id | -I id */
	 tempi = atoi(optarg) ;
	 if (!tempi) {
	    fprintf(stderr, "main:\t\t\tatoi(%s) failed. id must be an integer (0-65535)\n", optarg) ;
	    exit(1) ;
	 }
	 
	 conf.tunnel_id = (int16_t)tempi ;   
	 
	 
	 break ;
	 
       case 'i':
	 tempi = atoi(optarg) ; 
	 if (!tempi) {
	    fprintf(stderr, "main:\t\t\taoti(%s) failed. rid must be an integer (0-65535)\n", optarg) ;
	    exit(1) ;
	 }
	 
	 conf.tunnel_rid = tempi ;
	 
	 break ;
	 
	 
       case 'L': /* --listen | -L port */
	 conf.listen_port = atoi(optarg) ;
	 if (!conf.listen_port) {
	    fprintf(stderr, "main:\t\t\tatoi(%s) failed", optarg) ;
	    perror(" ") ;
	    exit(1) ;
	 }
	 conf.opmode = OPMODE_LISTEN ; 
	 
	 break ;
	 
       case 'l': /* --list | -l  (list ICMP types) */ /* FIXME: we should really store ICMP types and descriptions in a struct array */
	 fprintf(stderr, "ICMP types:\n"
			 "\tICMP_ECHO\n"
			 "\tICMP_ECHOREPLY\n"
			 "\tICMP_DEST_UNREACH\n"
			 "\tICMP_SOURCE_QUENCH\n"
			 "\tICMP_REDIRECT\n"
			 "\tICMP_ECHO\n"
       			 "\tICMP_TIME_EXCEEDED\n"
      			 "\tICMP_PARAMETERPROB\n"
     			 "\tICMP_TIMESTAMP\n"
      		 	 "\tICMP_TIMESTAMPREPLY\n"
		     	 "\tICMP_INFO_REQUEST\n" 
     			 "\tICMP_INFO_REPLY\n"
      			 "\tICMP_ADDRESS\n"
     			 "\tICMP_ADDRESSREPLY\n"
		   	 "\n") ;
		  exit(0) ;
		  break ;
	 
       case 'P':  /* --poll | -P  mode */
	 if (strcmp(optarg, "server")==0) 
	   conf.poll_mode = POLLMODE_SERVER ;
	 else if (strcmp(optarg, "client")==0)
	   conf.poll_mode = POLLMODE_CLIENT ;
	 else {
	    fprintf(stderr, "main:\t\t\tUnknown poll mode: '%s': Should be either 'server' or 'client'\n", optarg) ;
	    exit(1) ;
	 }
	 break ;
	 


					  
       case 'R': /* --recv_type | -R icmp_type */
	 tempi = icmptype_const_to_int(optarg) ;
	 if (tempi<0) {
	    fprintf(stderr, "main:\t\t\tUnknown ICMP type: %s (see -l)\n", optarg) ;
	    exit(1) ;
	 }
	 conf.recv_type = tempi ;
	 break ;
	 
       case 'S': /* --send_type | -S icmp_type */
	 tempi = icmptype_const_to_int(optarg) ;
	 if (tempi<0) {
	    fprintf(stderr, "main:\t\t\tUnknown ICMP type: %s (see -l)\n", optarg) ;
	    exit(1) ;
	 }
	 conf.send_type = tempi ;
	 break ; 
	 
       case 'V': /* --version | -V  */
	 print_version() ;
	 exit(0) ;
	 break ;
	 
       case '?':
	 print_usage(argv[0]) ;
	 exit(1) ; 
	 break ;
       case ':':
	 fprintf(stderr, "main:\t\t\tSHOOOO!\n") ;
	 break ;
	 
       default:
	 print_usage(argv[0]) ;
	 exit(1) ; 
	 
      }
      

       if (optind == (argc -1)) {
	  szhostname = (char *) malloc(strlen(argv[optind]) + 1) ;
	  sprintf(szhostname, "%s", argv[optind]) ;

	  	   
	  
	           to = (struct sockaddr_in *) &peer ;
	  
	           if (!(inet_aton(szhostname, &to->sin_addr))) {
		                  hp = gethostbyname(szhostname) ;
		                  if (!hp) {
				                    fprintf(stderr, "main:\t\t\tUnknown host: %s\n", szhostname) ;
				                    exit(1) ;
				  }
		      
		      
		                  to->sin_family = hp->h_addrtype ;
		      
		                  if (hp->h_length > (int)sizeof(to->sin_addr)) {
				                    hp->h_length = sizeof(to->sin_addr);
				  }
		                  memcpy(&to->sin_addr, hp->h_addr, hp->h_length);
		      
		      
		   }
	  	sprintf(conf.peer_ip, "%s", inet_ntoa(to->sin_addr)) ;

       }
   }
      

      
	    
	 

      
      
#else 
      
      fprintf(stderr, "main:\t\t\tWARNING: not using getopt_long(). \n") ;
   	if (argc>1) {
	   to = (struct sockaddr_in *) &peer ;
	   if (!(inet_aton(argv[1], &to->sin_addr)))
	     die("main:\tinet_aton() failed") ;
	} else {
	   to = (struct sockaddr_in *) &peer ;
	   inet_aton("127.0.0.1", &to->sin_addr) ;
	}

           if (argc>2) {
	                 port = atoi(argv[1]) ;
	   }
   
#endif

   if (strlen(conf.peer_ip)<(1*4)+3) {
      print_usage(argv[0]);
      exit(1) ;
   }
   

	 fprintf(stderr, "main:\t\t\tconf.\n"
		 "\t\t\t\tlisten_port=%i\n"
		 "\t\t\t\tforward_port=%i\n"
		 "\t\t\t\trecv_type=%s\n"
		 "\t\t\t\tsend_type=%s\n"
		 "\t\t\t\ttunnel_id=%i\n"
		 "\t\t\t\ttunnel_rid=%i\n"
		 "\t\t\t\tforward_host_ip=%s\n"
		 "\t\t\t\tpoll_mode=%i\n"
		 "\t\t\t\tpoll_interval=%i\n"
		 "\t\t\t;\n", 
		 conf.listen_port, 
		 conf.forward_port,
		 print_icmptype_const(conf.recv_type),
		 print_icmptype_const(conf.send_type),
		 conf.tunnel_id,
		 conf.tunnel_rid,
		 conf.forward_host_ip,
		 conf.poll_mode,
		 conf.poll_interval
		 ) ;
						      

   
	   
   s = create_socket(SOCK_STREAM) ;
   is = create_socket(SOCK_RAW) ;
   setup_icmp_socket(is, conf.recv_type, conf.send_type, conf.tunnel_id, conf.tunnel_rid) ;
   sockets[is].poll_mode = conf.poll_mode ;
   sockets[is].poll_interval = conf.poll_interval ; 


   switch(conf.opmode) {
    case OPMODE_LISTEN:			/* listen */
      bind_port(s, conf.listen_port) ;
      start_listening(s) ; 
      n = await_connection(s) ;
      break ;
      
    case OPMODE_FORWARD:		/* Forward */
      forward_connect(s, conf.forward_port, conf.forward_host_ip) ;
      n = s ;
      break ;
      
    default:
      fprintf(stderr, "main:\t\t\tUnknown opmode: %i\n", conf.opmode);
      exit(1) ;
      break ;
      
   }

     
   	
      
      tie_sockets(n, is) ; 


   	if (sockets[is].poll_mode == POLLMODE_SERVER) {
	   fprintf(stderr, "main:\t\t\tSending wake-up SPOLL to peer\n") ;
	   sockets[is].poll_key = POLLPROT_SPOLL ;
	   send_to_socket(is, NULL, 0) ;
	}
	   

   
   
	while (1) {
	
	   /* sockets 
	    * s = listen()ing socket
	    * n = accept()ed socket
	    * is = icmp socket 
	    */
	   


	   forward_socket_data(n, is) ;
	   

	   
	   
	}
   
   


   
   fprintf(stderr, "\n\n") ; 
   exit(0) ; 
   
}
   
   

int print_usage(const char *me) 
{
/*   print_version() ; */
   fprintf(stderr, "\n========================================================\n") ;

   fprintf(stderr, "\nUsage: %s [OPTIONS] ... HOST\n\n"
	        " Set up an icmp tunnel connection to HOST\n"
	        " --forward or --listen must be specified.\n"
	        " --send_type, --recv_type and --id are not mandatory, but preferred (defaults are bad).\n\n"
  		"\t-F | --forward host:port\tconnect to host:port and use it as endpoint\n"
   	      	"\t-L | --listen port\t\tlisten to local port, on connection use it as endpoint\n"
   		"\t-S | --send_type ICMP_type\tICMP type for outgoing packets (see -l) *\n"
   		"\t-R | --recv_type ICMP_type\tICMP type for incoming packets (see -l) *\n"
   		"\t-I | --id num\t\t\tID must not be the same on both peers. *\n"
	        "\t-i | --rid num\t\t\tRemote ID, the ID of peer. must not be the same on both peers. *\n"
	        "\t-P | --poll server|client\tAct as a Server- or Poll-Client. (see README for more information) *\n"
	        "\t-l | --list\t\t\tList ICMP types (for -R and -S) *\n"
	        "\t-h | --help\t\t\tShow this help message. *\n"
	   	"\t-e | --example\t\t\tShow example usage. *\n"
	        "\t-V | --version\t\t\tShow version and copyright information and exit *\n"
	        "\n(options marked with '*' are optional)\n"
	        "\n\nReport bugs to logic@nocrew.org\n"
   		"\n\n", me) ;
   
   return 0 ;
   

   

}

int print_example (const char *me)
{
   	fprintf(stdout, 
		"\nExample:\n"
                " This is an example telnet tunnel betwen box.blocked.net and box.free.net.\n"
                "   On box.blocked.net run:\n"
                "    %s -L 10023 --send_type ICMP_ECHO --recv_type ICMP_ECHOREPLY --id 1024 --rid 4096 box.free.net\n\n"
                "   And on box.free.net run:\n"
                "    %s -F box.free.net:23 --send_type ICMP_ECHOREPLY --recv_type ICMP_ECHO --id 4096 --rid 1024 box.blocked.net\n\n"
                "   Finally, on box.blocked.net run:\n"
                "    telnet box.blocked.net 10023\n\n"
		" Telnet data should now be forwarded betwen box.free.net:23 and box.blocked.net:10023 over ICMP_ECHO* frames\n"
		, me, me) ;

   return 0  ;
}

   
   
      


int print_version()
{
   printf("icmptunnel v%i.%i.%i  -  Tunnels data over ICMP frames.\n", VERSION_MAJOR, VERSION_MINOR, VERSION_REV) ;
   printf("Copyright 1999 Magnus Lundstrm (logic@nocrew.org) (See LICENSE for details)\n") ;
   
   return(0) ;
   
}

